@@ -5,12 +5,12 @@ from gastimator import gastimator
 from astropy.io import fits
 import matplotlib.pyplot as plt
 from gastimator import corner_plot, priors
-from astropy import wcs
+#from astropy import wcs
 from scipy.optimize import Bounds
 from scipy.optimize import minimize
 import matplotlib
 import warnings
-warnings.filterwarnings('ignore', category=wcs.FITSFixedWarning, append=True)
+#warnings.filterwarnings('ignore', category=wcs.FITSFixedWarning, append=True)
 import matplotlib.gridspec as gridspec
 from matplotlib.offsetbox import AnchoredText
 import astropy.units as u
@@ -22,6 +22,9 @@ from kinms_fitter.transformClouds import transformClouds
 from kinms_fitter.velocity_profs import velocity_profs      
 from kinms_fitter.prior_funcs import prior_funcs
 import time
+from spectral_cube import SpectralCube
+from spectral_cube.utils import SpectralCubeWarning
+warnings.filterwarnings(action='ignore', category=SpectralCubeWarning, append=True)
          
 class kinms_fitter:
     def __init__(self,filename,spatial_trim=None,spectral_trim=None,linefree_chans=[1,5]):
@@ -35,7 +38,7 @@ class kinms_fitter:
         self.spatial_trim=spatial_trim
         self.clip_cube()
         
-        self.wcs=wcs.WCS(self.hdr)
+        #self.wcs=wcs.WCS(self.hdr)
         self.xc_img=np.nanmedian(self.x1)
         self.yc_img=np.nanmedian(self.y1)
         self.xc_guess=np.nanmedian(self.x1)
@@ -74,8 +77,15 @@ class kinms_fitter:
         self.timetaken=0
         self.initial_guesses=None
         self.chi2_var_correct=True
+        self.chi2_correct_fac=None
         self.mask_sum=0
         self.labels=None
+        self.pa_prior,self.xc_prior,self.yc_prior,self.vsys_prior,self.inc_prior,self.totflux_prior,self.velDisp_prior = None,None,None,None,None,None,None
+        self.spectralcube=None
+        self.interactive=True
+        self.show_plots=True
+        self.output_cube_fileroot=False
+        
         self.tolerance=0.1 ## tolerance for simple fit. Smaller numbers are more stringent (longer runtime)
         try:
             self.objname=self.hdr['OBJECT']
@@ -92,60 +102,99 @@ class kinms_fitter:
         cax.xaxis.set_label_position('top')
         return cb
             
+    # def read_in_a_cube(self,path):
+    #     hdulist=fits.open(path)
+    #     hdr=hdulist[0].header
+    #     cube = np.squeeze(hdulist[0].data.T) #squeeze to remove singular stokes axis if present
+    #     cube[np.isfinite(cube) == False] = 0.0
+    #
+    #     try:
+    #         if hdr['CASAMBM']:
+    #             beamtab = hdulist[1].data
+    #     except:
+    #         beamtab=None
+    #
+    #     return cube, hdr, beamtab
     def read_in_a_cube(self,path):
-        hdulist=fits.open(path)
-        hdr=hdulist[0].header
-        cube = np.squeeze(hdulist[0].data.T) #squeeze to remove singular stokes axis if present
+        self.spectralcube=SpectralCube.read(path).with_spectral_unit(u.km/u.s, velocity_convention='radio')#, rest_value=self.restfreq)
+
+        hdr=self.spectralcube.header
+        cube = np.squeeze(self.spectralcube.filled_data[:,:,:].T).value #squeeze to remove singular stokes axis if present
         cube[np.isfinite(cube) == False] = 0.0
         
         try:
-            if hdr['CASAMBM']:
-                beamtab = hdulist[1].data
+            beamtab=self.spectralcube.beam
         except:
-            beamtab=None
-            
+            try:
+                beamtab=self.spectralcube.beams[np.floor(self.spectralcube.beams.size/2).astype(int)]
+            except:
+            #try flipping them
+                try:
+                    beamvals=[self.spectralcube.header['bmaj'],self.spectralcube.header['bmin']]
+                    beamtab=Beam(major=np.max(beamvals)*u.deg,minor=np.min(beamvals)*u.deg,pa=self.spectralcube.header['bpa']*u.deg)
+                except:
+                    beamtab=False
+ 
         return cube, hdr, beamtab
         
     def get_header_coord_arrays(self,hdr):
-        self.wcs=wcs.WCS(hdr)
-        self.wcs=self.wcs.sub(['longitude','latitude','spectral'])
-        maxsize=np.max([hdr['NAXIS1'],hdr['NAXIS2'],hdr['NAXIS3']])
 
-        xp,yp=np.meshgrid(np.arange(0,maxsize),np.arange(0,maxsize))
-        zp = xp.copy()
-
-        x,y,spectral = self.wcs.all_pix2world(xp,yp,zp, 0)
-        
-        
-
-        x1=np.median(x[0:hdr['NAXIS1'],0:hdr['NAXIS2']],0)
-        y1=np.median(y[0:hdr['NAXIS1'],0:hdr['NAXIS2']],1)
-        spectral1=spectral[0,0:hdr['NAXIS3']]
-
-        if (hdr['CTYPE3'] =='VRAD') or (hdr['CTYPE3'] =='VELO-LSR') or (hdr['CTYPE3'] =='VOPT') or (hdr['CTYPE3'] =='FELO-HEL'):
-            v1=spectral1
-            try:
-                if hdr['CUNIT3']=='m/s':
-                     v1/=1e3
-                     
-            except:
-                     v1/=1e3
-
-
-        else:
-           f1=spectral1*u.Hz
-           try:
-               self.restfreq = hdr['RESTFRQ']*u.Hz
-           except:
-               self.restfreq = hdr['RESTFREQ']*u.Hz
-           v1=f1.to(u.km/u.s, equivalencies=u.doppler_radio(self.restfreq))
-           v1=v1.value
+        y,x=self.spectralcube.spatial_coordinate_map
 
+        x1=np.median(x[0:hdr['NAXIS2'],0:hdr['NAXIS1']],0).value
+        y1=np.median(y[0:hdr['NAXIS2'],0:hdr['NAXIS1']],1).value
+        v1=self.spectralcube.spectral_axis.value
 
         cd3= np.median(np.diff(v1))
-        cd1= np.median(np.diff(y1))
+        cd1= np.median(np.diff(x1))
         
+        if np.any(np.diff(x1) > 359):
+            # we have RA=0 wrap issue
+            x1[x1 > 180]-=360
+            
+            
         return x1,y1,v1,np.abs(cd1*3600),cd3
+            
+    # def get_header_coord_arrays(self,hdr):
+    #     self.wcs=wcs.WCS(hdr)
+    #     self.wcs=self.wcs.sub(['longitude','latitude','spectral'])
+    #     maxsize=np.max([hdr['NAXIS1'],hdr['NAXIS2'],hdr['NAXIS3']])
+    #
+    #     xp,yp=np.meshgrid(np.arange(0,maxsize),np.arange(0,maxsize))
+    #     zp = xp.copy()
+    #
+    #     x,y,spectral = self.wcs.all_pix2world(xp,yp,zp, 0)
+    #
+    #
+    #
+    #     x1=np.median(x[0:hdr['NAXIS1'],0:hdr['NAXIS2']],0)
+    #     y1=np.median(y[0:hdr['NAXIS1'],0:hdr['NAXIS2']],1)
+    #     spectral1=spectral[0,0:hdr['NAXIS3']]
+    #
+    #     if (hdr['CTYPE3'] =='VRAD') or (hdr['CTYPE3'] =='VELO-LSR') or (hdr['CTYPE3'] =='VOPT') or (hdr['CTYPE3'] =='FELO-HEL'):
+    #         v1=spectral1
+    #         try:
+    #             if hdr['CUNIT3']=='m/s':
+    #                  v1/=1e3
+    #
+    #         except:
+    #                  v1/=1e3
+    #
+    #
+    #     else:
+    #        f1=spectral1*u.Hz
+    #        try:
+    #            self.restfreq = hdr['RESTFRQ']*u.Hz
+    #        except:
+    #            self.restfreq = hdr['RESTFREQ']*u.Hz
+    #        v1=f1.to(u.km/u.s, equivalencies=u.doppler_radio(self.restfreq))
+    #        v1=v1.value
+    #
+    #
+    #     cd3= np.median(np.diff(v1))
+    #     cd1= np.median(np.diff(y1))
+    #     breakpoint()
+    #     return x1,y1,v1,np.abs(cd1*3600),cd3
            
     def rms_estimate(self,cube,chanstart,chanend):
         quarterx=np.array(self.x1.size/4.).astype(np.int)
@@ -200,9 +249,9 @@ class kinms_fitter:
         datacube,hdr,beamtab = self.read_in_a_cube(cube)
         
         try:
-           self.bmaj=np.median(beamtab['BMAJ'])
-           self.bmin=np.median(beamtab['BMIN'])
-           self.bpa=np.median(beamtab['BPA'])
+           self.bmaj= beamtab.major.to(u.arcsec).value
+           self.bmin= beamtab.minor.to(u.arcsec).value
+           self.bpa=beamtab.pa.to(u.degree).value
         except:     
            try:
                self.bmaj=hdr['BMAJ']*3600.
@@ -241,7 +290,7 @@ class kinms_fitter:
         maximums=np.array([self.pa_range[1],self.xcent_range[1],self.ycent_range[1],self.vsys_range[1],self.inc_range[1],self.totflux_range[1],self.velDisp_range[1]])
         labels=np.array(["PA","Xc","Yc","Vsys","inc","totflux","veldisp"])
         fixed= minimums == maximums
-        priors=np.resize(None,fixed.size)
+        priors=np.array([self.pa_prior,self.xc_prior,self.yc_prior,self.vsys_prior,self.inc_prior,self.totflux_prior,self.velDisp_prior])#np.resize(None,fixed.size)
         precision=(maximums-minimums)/10.        
 
         
@@ -284,39 +333,38 @@ class kinms_fitter:
         
             
         
-    def model(self,param):
-        pa=param[0]
-        xc=param[1]
-        yc=param[2]
-        vsys=param[3]
-        inc=param[4]
-        totflux=param[5]
-        veldisp=param[6]
-        phasecen=[(xc-self.xc_img)*3600.,(yc-self.yc_img)*3600.]
-        
-        
-        vrad=velocity_profs.eval(self.vel_profile,self.sbRad,param[7+self.n_sbvars:7+self.n_velvars+self.n_sbvars],inc=inc)
-        
-        if self.n_radmotionvars >0:
-            radmotion=self.radial_motion[0](self.sbRad,param[7+self.n_velvars+self.n_sbvars:])
-        else:
-            radmotion=None
-        
-        if len(self.skySampClouds) >0:
-            inClouds=transformClouds(self.skySampClouds[:,0:3],posAng = pa,inc = inc,cent = phasecen)
-            fluxclouds=self.skySampClouds[:,3]
-            sbprof=None
-        else:
-            sbprof=sb_profs.eval(self.sb_profile,self.sbRad,param[7:7+self.n_sbvars])
-            inClouds=[]
-            fluxclouds=None
-        
-        #breakpoint()
-        return KinMS(self.x1.size*self.cellsize,self.y1.size*self.cellsize,self.v1.size*self.dv,self.cellsize,self.dv,\
-                 [self.bmaj,self.bmin,self.bpa],inc,sbProf=sbprof,sbRad=self.sbRad,velRad=self.sbRad,velProf=vrad,gasSigma=veldisp,\
-                 intFlux=totflux,posAng=pa,fixSeed=True,vOffset=vsys - self.vsys_mid,phaseCent=phasecen,nSamps=self.nSamps,vSys=vsys,radial_motion_func=radmotion,inClouds=inClouds,flux_clouds=fluxclouds).model_cube()
-                
-    def model_simple(self,param):
+    # def model(self,param):
+    #     pa=param[0]
+    #     xc=param[1]
+    #     yc=param[2]
+    #     vsys=param[3]
+    #     inc=param[4]
+    #     totflux=param[5]
+    #     veldisp=param[6]
+    #     phasecen=[(xc-self.xc_img)*3600.,(yc-self.yc_img)*3600.]
+    #
+    #
+    #     vrad=velocity_profs.eval(self.vel_profile,self.sbRad,param[7+self.n_sbvars:7+self.n_velvars+self.n_sbvars],inc=inc)
+    #
+    #     if self.n_radmotionvars >0:
+    #         radmotion=self.radial_motion[0](self.sbRad,param[7+self.n_velvars+self.n_sbvars:])
+    #     else:
+    #         radmotion=None
+    #
+    #     if len(self.skySampClouds) >0:
+    #         inClouds=transformClouds(self.skySampClouds[:,0:3],posAng = pa,inc = inc,cent = phasecen)
+    #         sbprof=None
+    #         myargs={'vPhaseCent': phasecen,'inClouds': inClouds, 'flux_clouds': self.skySampClouds[:,3]}
+    #     else:
+    #         sbprof=sb_profs.eval(self.sb_profile,self.sbRad,param[7:7+self.n_sbvars])
+    #         myargs={'phaseCent': phasecen}
+    #
+    #
+    #     return KinMS(self.x1.size*self.cellsize,self.y1.size*self.cellsize,self.v1.size*self.dv,self.cellsize,self.dv,\
+    #              [self.bmaj,self.bmin,self.bpa],inc,sbProf=sbprof,sbRad=self.sbRad,velRad=self.sbRad,velProf=vrad,gasSigma=veldisp,\
+    #              intFlux=totflux,posAng=pa,fixSeed=True,vOffset=vsys - self.vsys_mid,nSamps=self.nSamps,vSys=vsys,radial_motion_func=radmotion,**myargs).model_cube()
+    
+    def model_simple(self,param,fileName=''):
         pa=param[0]
         xc=param[1]
         yc=param[2]
@@ -324,15 +372,15 @@ class kinms_fitter:
         inc=param[4]
         totflux=param[5]
         veldisp=param[6]
+        phasecen=[xc,yc]
 
         if len(self.skySampClouds) >0:
-            inClouds=transformClouds(self.skySampClouds[:,0:3],posAng = pa,inc = inc,cent = [xc,yc])
-            fluxclouds=self.skySampClouds[:,3]
+            inClouds=transformClouds(self.skySampClouds[:,0:3],posAng = pa,inc = inc,cent = phasecen)
             sbprof=None
+            myargs={'vPhaseCent': phasecen,'inClouds': inClouds, 'flux_clouds': self.skySampClouds[:,3]}
         else:
             sbprof=sb_profs.eval(self.sb_profile,self.sbRad,param[7:7+self.n_sbvars])
-            inClouds=None
-            fluxclouds=None
+            myargs={'phaseCent': phasecen}
         
         vrad=velocity_profs.eval(self.vel_profile,self.sbRad,param[7+self.n_sbvars:],inc=inc)
         
@@ -341,19 +389,20 @@ class kinms_fitter:
         else:
             radmotion=None
         
+        
         return KinMS(self.x1.size*self.cellsize,self.y1.size*self.cellsize,self.v1.size*self.dv,self.cellsize,self.dv,\
                  [self.bmaj,self.bmin,self.bpa],inc,sbProf=sbprof,sbRad=self.sbRad,velRad=self.sbRad,velProf=vrad,gasSigma=veldisp,\
-                 intFlux=totflux,posAng=pa,fixSeed=True,vOffset=vsys - self.vsys_mid,phaseCent=[xc,yc],nSamps=self.nSamps,vSys=vsys,radial_motion_func=radmotion,inClouds=inClouds,flux_clouds=fluxclouds).model_cube()
+                 intFlux=totflux,posAng=pa,fixSeed=True,vOffset=vsys - self.vsys_mid,nSamps=self.nSamps,vSys=vsys,radial_motion_func=radmotion,ra=self.xc_img,dec=self.yc_img,fileName=fileName,**myargs).model_cube()
         
             
         
     def mcmc_fit(self,initial_guesses,labels,minimums,maximums,fixed,priors,precision):
-        imx=self.xc_img
-        imy=self.yc_img    
-        initial_guesses[1]=(initial_guesses[1]-self.xc_img)*3600.
-        initial_guesses[2]=(initial_guesses[2]-self.yc_img)*3600.
-        self.xc_img=0
-        self.yc_img=0
+        # imx=self.xc_img
+        # imy=self.yc_img
+        # initial_guesses[1]=(initial_guesses[1]-self.xc_img)*3600.
+        # initial_guesses[2]=(initial_guesses[2]-self.yc_img)*3600.
+        # self.xc_img=0
+        # self.yc_img=0
         minimums[1:3]=(self.bmaj*-3)
         maximums[1:3]=(self.bmaj*3)
         precision[1:3]=(maximums[1:3]-minimums[1:3])*0.1
@@ -380,24 +429,25 @@ class kinms_fitter:
             print("Parameters Fixed:",labels[mcmc.fixed]) 
     
         
-        if self.chi2_var_correct:
-            correction_factor=((2*self.mask_sum)**0.25)
-        else:
-            correction_factor=1
+        if self.chi2_var_correct & (self.chi2_correct_fac == None):
+           self.chi2_correct_fac=(2*(self.mask_sum**0.25))
+
+        if not self.chi2_var_correct:
+            self.chi2_correct_fac=1
             
         if not self.silent: 
             if self.chi2_var_correct:
-                print("Correction for chi-sqr variance applied:",correction_factor)
+                print("Correction for chi-sqr variance applied:",self.chi2_correct_fac)
             else:
                 print("Correction for chi-sqr variance not applied")    
         
-        outputvalue, outputll= mcmc.run(self.cube,self.error*correction_factor,self.niters,nchains=1,plot=False)
-        
-        outputvalue[1]=imx+(outputvalue[1]/3600.)
-        outputvalue[2]=imy+(outputvalue[2]/3600.)
-        self.xc_img=imx
-        self.yc_img=imy
+        outputvalue, outputll= mcmc.run(self.cube,self.error*self.chi2_correct_fac,self.niters,nchains=1,plot=False)
         
+        # outputvalue[1]=imx+(outputvalue[1]/3600.)
+        # outputvalue[2]=imy+(outputvalue[2]/3600.)
+        # self.xc_img=imx
+        # self.yc_img=imy
+
         bestvals=np.median(outputvalue,1)    
         besterrs=np.std(outputvalue,1)
         
@@ -426,12 +476,12 @@ class kinms_fitter:
 
             
     def simple_fit(self,initial_guesses,labels,minimums,maximums,fixed):
-        imx=self.xc_img
-        imy=self.yc_img    
-        initial_guesses[1]=(initial_guesses[1]-self.xc_img)*3600.
-        initial_guesses[2]=(initial_guesses[2]-self.yc_img)*3600.
-        self.xc_img=0
-        self.yc_img=0
+        # imx=self.xc_img
+        # imy=self.yc_img
+        # initial_guesses[1]=(initial_guesses[1]-self.xc_img)*3600.
+        # initial_guesses[2]=(initial_guesses[2]-self.yc_img)*3600.
+        # self.xc_img=0
+        # self.yc_img=0
         minimums[1:3]=(self.bmaj*-3)
         maximums[1:3]=(self.bmaj*3)
         
@@ -440,24 +490,30 @@ class kinms_fitter:
         
         self.bounds = Bounds(minimums, maximums,keep_feasible=True)
 
-        res = minimize(self.simple_chi2, initial_guesses, args={'Nfeval':0},method ='Powell' ,bounds=self.bounds, options={'disp': True,'adaptive':True,'maxfev':self.niters,'ftol':self.tolerance}) 
+        res = minimize(self.simple_chi2, initial_guesses, args={'Nfeval':0},method ='Powell' ,bounds=self.bounds, options={'disp': True,'adaptive':True,'maxfev':self.niters,'maxiter':self.niters,'ftol':self.tolerance}) 
         
         results=res.x
-        results[1]=imx+(results[1]/3600.)
-        results[2]=imy+(results[2]/3600.)
-        self.xc_img=imx
-        self.yc_img=imy
+        # results[1]=imx+(results[1]/3600.)
+        # results[2]=imy+(results[2]/3600.)
+        # self.xc_img=imx
+        # self.yc_img=imy
             
         return results 
 
-    def plot(self,overcube=None,savepath=None,**kwargs):
+    def plot(self,block=True,overcube=None,savepath=None,**kwargs):
         pl=KinMS_plotter(self.cube.copy(), self.x1.size*self.cellsize,self.y1.size*self.cellsize,self.v1.size*self.dv,self.cellsize,self.dv,[self.bmaj,self.bmin,self.bpa], posang=self.pa_guess,overcube=overcube,rms=self.rms,savepath=savepath,savename=self.pdf_rootname,**kwargs)
-        pl.makeplots()
+        pl.makeplots(block=block,plot2screen=self.show_plots)
         self.mask_sum=pl.mask.sum()
         
     def run(self,method='mcmc',justplot=False,**kwargs):
         self.bincentroids=np.arange(0,self.nrings)*self.bmaj
         self.error=self.rms
+
+        self.xc_guess=(self.xc_guess-self.xc_img)*3600.
+        self.yc_guess=(self.yc_guess-self.yc_img)*3600.
+
+        
+        
         
         if np.any(self.sb_profile) == None:
             # default SB profile is a single exponential disc
@@ -484,8 +540,9 @@ class kinms_fitter:
         initial_guesses,labels,minimums,maximums,fixed, priors,precision = self.setup_params()
         self.labels=labels
         
+        
         t=time.time()
-        init_model=self.model(initial_guesses)
+        init_model=self.model_simple(initial_guesses)
         self.timetaken=(time.time()-t)
         
         
@@ -494,9 +551,9 @@ class kinms_fitter:
             print("==============   Welcome to KinMS_fitter!   ==============")
             print(self.logo())
             print("==========================================================")
-            print("One model evaulation takes {:.2f} seconds".format(self.timetaken))
+            print("One model evaluation takes {:.2f} seconds".format(self.timetaken))
         
-        self.plot(overcube=init_model,**kwargs)
+        self.plot(overcube=init_model,block=self.interactive,**kwargs)
 
         if justplot:
             return 1,1,1,1,1
@@ -524,22 +581,37 @@ class kinms_fitter:
                 if not self.silent: 
                     print("MCMC fitting process took {:.2f} seconds".format((time.time()-t)))
             self.pa_guess=bestvals[0]
-            best_model=self.model(bestvals)
+            
+            if self.output_cube_fileroot != False:
+                fileName=self.output_cube_fileroot
+            else:
+                fileName=''
+            best_model=self.model_simple(bestvals,fileName=fileName)
             if self.pdf:
                 savepath="./"
             else:
                 savepath=None
                 
-            self.plot(overcube=best_model,savepath=savepath,**kwargs)
+            self.plot(overcube=best_model,savepath=savepath,block=self.interactive,**kwargs)
             
             
-            if ((method=='mcmc') or (method=='both')) and self.show_corner:   
-                fig=corner_plot.corner_plot(outputvalue[~fixed,:].T,like=outputll,\
+            if ((method=='mcmc') or (method=='both')) and self.show_corner:
+                if np.any(np.std(outputvalue[~fixed,:],1)==0): 
+                    print('Some parameters had no accepted guesses. Skipping corner plot. Trying increasing niters.')
+                else:
+                    fig=corner_plot.corner_plot(outputvalue[~fixed,:].T,like=outputll,\
                                         quantiles=[0.16, 0.5, 0.84],labels=self.labels[~fixed],verbose=False)
-                if self.pdf:
+                    if self.pdf:
+                        plt.savefig(self.pdf_rootname+"_MCMCcornerplot.pdf")
+                    if self.show_plots:    
+                        if self.interactive==False:
+                            plt.draw()
+                            plt.pause(1e-6)
+                        else:
+                            plt.show()
                     
-                    plt.savefig(self.pdf_rootname+"_MCMCcornerplot.pdf")
-                plt.show()                        
+            bestvals[1]=(bestvals[1]/3600.)+self.xc_img
+            bestvals[2]=(bestvals[2]/3600.)+self.yc_img
                                         
         
             return bestvals, besterrs, outputvalue, outputll, fixed
